# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>

from xlms_tools.parsers import parsebiopystructure
import numpy as np

atomradius={'CA': 1.90,
            'C': 1.88,
            'N': 1.63,
            'O': 1.48,
            'S': 1.78,
            'H': 1.2,
            'P': 1.87,
            'CB': 1.96,
            'NE': 1.63,
            'FE': 0.74,
            'default': 1.8, 
            'OX': 1.48, 
            'HX': 1.2}

def selectfromatomhash(hashtab, atomname):
    if atomname[0] == 'C':
        if atomname == 'CA':
            return hashtab['CA']
        elif atomname == 'C':
            return hashtab['C']
        else:
            return hashtab['CB']
    elif atomname[0] == 'O':
        if atomname == 'O':
            return hashtab['O']
        else:
            return hashtab['OX']
    elif atomname[0] == 'N':
        if atomname == 'N':
            return hashtab['N']
        else:
            return hashtab['NE']
    elif atomname[0] == 'S':
        return hashtab['S']
    elif atomname[0] == 'P':
        return hashtab['P']
    elif atomname[0] == 'H':
        return hashtab['H']
    elif atomname[0:2] == 'FE':
        return hashtab['FE']
    else:
        return hashtab['default']

def computedepth(biopdbstruct, scale=0.5, proberadius=2.5, margin=2.5, return_indices=False, verbose=True):
    import time
    
    if verbose:
        print ("-- Building space-filling model: ", end="")
        
    start = time.time()
    atomlist, coords = listpdbatoms(biopdbstruct)
    atommasks = sphericalatommasks(scale, proberadius)
    objectbox, translatedatoms, scale = spacefillingmodel(coords, atomlist, atommasks, proberadius, scale, margin)
    
    if verbose:
        print (f'{time.time()-start:.3f}s')    
        
    start = time.time()
    edt, nearestwaters, image = edt_scipy(objectbox,scale)
    depths, residues = getresiduedepths(edt, translatedatoms, atomlist, proberadius)
    
    if verbose:
        print (f"-- Calculating residue depths: {time.time()-start:.3f}s")
        
    if return_indices == False:
        return depths, residues
    else:
        return depths, residues, edt, nearestwaters, image, atomlist, translatedatoms
    
# atomlist = listpdbatoms(biopdbstruct)
def listpdbatoms(biopdbstruct):
    atomlst = []
    coordslst = []
    for chain in biopdbstruct[0]:   # get only first model in structure
        for res in chain:
            for atom in res:
                if not atomisH(atom):
                    if (res.id[0][0:2] != 'H_') and (res.id[0][0:1] != 'W'):     # if not heteroatom
                        atomlst.append((chain.id, res.id[1], atom.name, res.resname, selectfromatomhash(atomradius, atom.name)))
                        coordslst.append(atom.coord)
    coords = np.array(coordslst)
    return atomlst, coords


def atomisH(biopdbatom):
    if biopdbatom.element == 'H':
        return True
    if biopdbatom.name[0] == 'H':
        return True
    return False

# atommasks = sphericalatommasks(scale, proberadius)
def sphericalatommasks(scale, proberadius):
    
    mask = {}   # voxels with atom 1, background voxels 0
    for atom in atomradius.keys():
        tradius = (atomradius[atom]+proberadius)*scale + 0.5
        sradius = tradius*tradius
        widxz = int(tradius) + 1
        depty = []
        indx = 0
        for j in range(widxz):
            for k in range(widxz):
                txz = j*j + k*k
                if (txz > sradius):
                    depty.append(-1)
                else:
                    tdept=np.sqrt(sradius-txz)
                    depty.append(int(tdept))
        mask[atom] = makesphere(widxz, depty)        
    return mask

def makesphere(widxz, depty):
    rad = widxz - 1
    dim = 2*rad + 1
    box = np.zeros((dim, dim, dim))
    
    nind = 0
    for i in range(widxz):
        for j in range(widxz):
            if (depty[nind] != -1):
                for ii in range(-1, 2):
                    for jj in range(-1, 2):
                        for kk in range(-1, 2):
                            if(ii!=0 and jj!=0 and kk!=0):
                                mi = ii*i
                                mk = kk*j
                                for k in range(depty[nind]+1):
                                    mj = k*jj
                                    si = rad+mi
                                    sj = rad+mj
                                    sk = rad+mk
                                    box[si][sj][sk] = 1
            nind += 1
    return box
    
# translatedatoms = scaleandtranslate(coords, scale, proberadius)
def scaleandtranslate(coord, scale, proberadius, margin):
    deltax = coord[:,0].min() - (proberadius+margin)
    deltay = coord[:,1].min() - (proberadius+margin)
    deltaz = coord[:,2].min() - (proberadius+margin)
    return (coord - np.array([deltax, deltay, deltaz]))*scale

#boxlenx, boxleny, boxlenz = computeboxsize(coords, scale, proberadius, margin)
def computeboxsize(coords, scale, proberadius, margin):
    boxlength = 128
    width = coords[:,0].max() - coords[:,0].min() + (proberadius+margin)*2
    length = coords[:,1].max() - coords[:,1].min() + (proberadius+margin)*2
    height = coords[:,2].max() - coords[:,2].min() + (proberadius+margin)*2

    boxlenx = int(np.ceil(scale*width) + 1)
    boxleny = int(np.ceil(scale*length) + 1)
    boxlenz = int(np.ceil(scale*height) + 1)

    return np.array([boxlenx, boxleny, boxlenz]), scale

#objectbox = spacefillingmodel(translatedatoms, atomlist, atommasks)
def spacefillingmodel(coords, atomlist, atommasks, proberadius, scale, margin):
    translatedatoms = scaleandtranslate(coords, scale, proberadius, margin)
    
    # compute box size
    boxlen, scale = computeboxsize(coords, scale, proberadius, margin)
    #print ("box: ", boxlen[0], boxlen[1], boxlen[2], "scale:", scale)
    objectbox=np.zeros((boxlen[0], boxlen[1], boxlen[2]))
    
    # for each atom, apply object and surface masks
    for i in range(len(atomlist)):
        atomtype = atomlist[i][2]
        obj = selectfromatomhash(atommasks, atomtype)
        #obj = selectatommask(atommasks, atomtype)
        radobj = (obj.shape[0]-1)/2
        objstart, objend = getboxbounds(translatedatoms[i,:], radobj)
        objectbox[objstart[0]:objend[0], objstart[1]:objend[1], objstart[2]:objend[2]] += obj

    objectbox[objectbox > 0] = 1

    return objectbox, translatedatoms, scale

# objstart, objend = getboxbounds(coord[i,:], radobj)
def getboxbounds(atomcoord, radius):
    rad = int(radius)
    centerx, centery, centerz = getvoxelposition(atomcoord)
    start = (centerx-rad, centery-rad, centerz-rad)
    end = (centerx+rad+1, centery+rad+1, centerz+rad+1)
    return start, end

def getvoxelposition(coords):
    vx = int(coords[0] + 0.5)
    vy = int(coords[1] + 0.5)
    vz = int(coords[2] + 0.5)
    
    return vx, vy, vz
        
# obj = selectatommask(objmasks, atomtype)
def selectatommask(masks, atomtype):
    return selectfromatomhash(masks, atomtype)

# depths = getresiduedepths(edt, translatedatoms, atomlist)
def getresiduedepths(edt, translatedatoms, atomlist, proberadius):
    depths = {}
    residues = {}
    for i in range(len(atomlist)):
        key = atomlist[i][0]+':'+str(atomlist[i][1])
        if key not in depths:
            depths[key] = []
            residues[key] = atomlist[i][3]
        x, y, z = getvoxelposition(translatedatoms[i])
        
        # check if minimum depth met
        if edt[x,y,z] - proberadius < atomlist[i][4]:
            curdep = atomlist[i][4] + proberadius
        else:
            curdep = edt[x,y,z]
             
        depths[key].append(curdep)
    
    for i in depths.keys():
        depths[i] = sum(depths[i])/len(depths[i])

    return depths, residues

def edt_scipy(image, scale):
    from scipy.ndimage import distance_transform_edt
    
    image, bg = peeloffouter(image)
    depths, nearestwater = distance_transform_edt(image, return_indices=True)
    depths[bg == -1] = -1    
    depths /= scale
    return depths, nearestwater, image

def peeloffouter(img):
    newimg = np.copy(img)
    bg = np.zeros((img.shape[0], img.shape[1], img.shape[2]))
    bg[img == 0] = -1
    
    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            for k in range(img.shape[2]):
                if img[i,j,k] > 0:  #is object pixel
                    istart = max(0, i-1)
                    jstart = max(0, j-1)
                    kstart = max(0, k-1)
                    iend = min(img.shape[0], i+2)
                    jend = min(img.shape[1], j+2)
                    kend = min(img.shape[2], k+2)
                    total = (iend - istart)*(jend-jstart)*(kend-kstart)
                    if np.sum(img[istart:iend, jstart:jend, kstart:kend]) < total:
                        newimg[i,j,k] = 0
    newimg -= bg
    return newimg, bg



# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>

from Bio.PDB import MMCIFParser, PDBParser
import numpy as np

def parsebiopystructure(infile):
    parser = None
    if infile[-3:].lower() == 'pdb':
        parser = PDBParser(QUIET=True)
    elif infile[-3:].lower() == 'cif':
        parser = MMCIFParser(QUIET=True)
    else:
        print (f'File suffix {infile[-3:]} not recognized: must either be pdb or cif')
        return None
    structure = parser.get_structure(infile[:-4], infile)
    return structure
    
def extractCAcoordsfrompdb(pdbfile):
    structure = parsebiopystructure(pdbfile)
    CAcoords = np.zeros(3)
    residues = structure.get_residues()
    nres = 0
    for res in residues:
        if res.has_id("CA"):
            atom = np.asarray(res["CA"].get_coord())
            CAcoords = np.concatenate((CAcoords, atom), axis=0)
            nres += 1        
    CAcoords = np.reshape(CAcoords[3:], (nres,3))
    return CAcoords

def maxCAdistfromstructure(structure):
    CAcoords = np.zeros(3)
    residues = structure.get_residues()
    nres = 0
    for res in residues:
        if res.has_id("CA"):
            atom = np.asarray(res["CA"].get_coord())
            CAcoords = np.concatenate((CAcoords, atom), axis=0)
            nres += 1        
    CAcoords = np.reshape(CAcoords[3:], (nres,3))
    distmat = distancematrix(CAcoords, CAcoords)
    return distmat.max()
    
    
def distancematrix(a, b):
    distmat = np.linalg.norm(a[:, None, :] - b[None, :, :], axis=-1)
    return distmat



# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>

from xlms_tools.depth import computedepth, getvoxelposition
from xlms_tools.parsers import parsebiopystructure
from os.path import splitext
from math import sqrt, ceil
import numpy as np

# GLOBAL VARIABLES
image = None
nearestwaters = None
candidatepaths = None
neighbor_offsets = [(dx, dy, dz) for dx in (-1, 0, 1) for dy in (-1, 0, 1) for dz in (-1, 0, 1) if (dx, dy, dz) != (0, 0, 0)]
backbone = ['N', 'CA', 'C', 'O']

def outputSASDs(pdbfile, linker='BS3/DSS', outfile=None):
    if outfile == None:
        outfile = splitext(pdbfile)[0] + '_sasds.txt'
    structure = parsebiopystructure(pdbfile)
    sasds = computeSASDs(structure, linker=linker)

    # print SASDs
    for atomA, atomB, sasdAB in sasds:
        print(atomA, atomB, sasdAB)
        
def computeSASDs(biopdbstruct, linker='BS3/DSS', maxdepth=6.25):
    global image
    global nearestwaters
    
    taggedresidues = {'BS3/DSS':['LYS'],
                        'DSBSO':['LYS'],
                        'PHOX':['LYS']
                    }
                    
    depths, residues, edt, nearestwaters, image, atomlist, translatedatoms = computedepth(biopdbstruct, return_indices=True)
    
    # note voxel locations of taggable Calphas
    allatoms = [(atomlist[i], translatedatoms[i,:]) for i in range(len(atomlist))]
    taggedatoms = [i for i in allatoms if i[0][3] in taggedresidues[linker] and depths[f'{i[0][0]}:{i[0][1]}'] <= maxdepth]
    taggedCAs = [i for i in taggedatoms if i[0][2] == 'CA']
    
    # find sidechain ends of tagged residues
    taggedSCends = []
    for ca in taggedCAs:
        curratoms = [i for i in taggedatoms if (i[0][0] == ca[0][0]) and (i[0][1] == ca[0][1]) and i[0][2] not in backbone]
        sortedatoms = sorted(curratoms, key = lambda x: euclideandistance(x[1], ca[1]))
        taggedSCends.append(sortedatoms[-1])
        
    # represent outershell as graph and compute all shortest paths between voxels
    distmatrix, voxelindex = graphshortestpaths(image, 0)
    
    # compute SASD for each pair of taggable Calphas
    sasds = []

    for i in range(len(taggedSCends)-1):
        voxi = getvoxelposition(taggedSCends[i][1])
        surfvoxi = watervoxel(voxi, nearestwaters)
        for j in range(i+1, len(taggedSCends)):
            voxj = getvoxelposition(taggedSCends[j][1])
            surfvoxj = watervoxel(voxj, nearestwaters)
            dist = distmatrix[voxelindex[surfvoxi], voxelindex[surfvoxj]]
            totalsasd = edt[voxi] + edt[voxj] + \
                        euclideandistance(voxi, taggedCAs[i][1]) + \
                        euclideandistance(voxj, taggedCAs[j][1]) + \
                        dist

            
            sasds.append((taggedCAs[i][0], taggedCAs[j][0], totalsasd))
    
#    for atom in allatoms:
#        chainid = atom[0][0]
#        resnum = atom[0][1]
#        atomname = atom[0][2]
#        resname = atom[0][3]
#        coords = atom[1]
    
    
#    for i in range(len(atomlist)):
#        chainid = atomlist[i][0]
#        resnum = atomlist[i][1]
#        atomname = atomlist[i][2]
#        resname = atomlist[i][3]
        ## reference for tuple format: atomlst.append((chain.id, res.id[1], atom.name, res.resname, selectfromatomhash(atomradius, atom.name)))
        
#        if (resname in taggedresidues[linker]):
#            key = f'{chainid}:{resnum}'
#            if key not in taggedatoms:
#                taggedatoms[key] = []
#            if atomname == 'CA':  # if CA atom found
#                if depths[key] <= maxdepth: # if within maximum residue depth
#                    taggedCAs.append((key, atomlist[i]))
#            taggedatoms[key].append(atomlist[i])
        
#        for key, atom in taggedCAs:
            
                #vx, vy, vz = getvoxelposition(translatedatoms[i,:])
                #taggedvoxels.append((vx, vy, vz))
                
    
    
    
#    for i in range(len(taggedvoxels)-1):
#    for i in range(1):
#        surfvoxi = watervoxel(taggedvoxels[i], nearestwaters)
#        for j in range(i+1, len(taggedvoxels)):
#        for j in range(i+1, 3):
#            candidatepaths = []
#            surfvoxj = watervoxel(taggedvoxels[j], nearestwaters)
#            if straightlinepathpossible(surfvoxi, surfvoxj):
#                dist = euclideandistance(surfvoxi, surfvoxj)
#            else:
                #crosslinkpathing([surfvoxi], surfvoxj)
                #dist = min([totaldistance(i) for i in candidatepaths])
#                dist = distmatrix[voxelindex[surfvoxi], voxelindex[surfvoxj]]

#            dist = distmatrix[voxelindex[surfvoxi], voxelindex[surfvoxj]]
#            totalsasd = edt[taggedvoxels[i][0], taggedvoxels[i][1], taggedvoxels[i][2]] + \
#                        edt[taggedvoxels[j][0], taggedvoxels[j][1], taggedvoxels[j][2]] + \
#                        dist
#            print (surfvoxi, edt[taggedvoxels[i][0], taggedvoxels[i][1], taggedvoxels[i][2]])
#            print (surfvoxj, edt[taggedvoxels[j][0], taggedvoxels[j][1], taggedvoxels[j][2]])
#            print (euclideandistance(surfvoxi, surfvoxj))
#            totalsasd = edt[taggedvoxels[i][0], taggedvoxels[i][1], taggedvoxels[i][2]] + edt[taggedvoxels[j][0], taggedvoxels[j][1], taggedvoxels[j][2]] + euclideandistance(surfvoxi, surfvoxj)

            
#            sasds.append((taggedCAs[i], taggedCAs[j], totalsasd))
    return sasds

def euclideandistance(voxA, voxB):
    return sqrt((voxA[0]-voxB[0])**2 + (voxA[1]-voxB[1])**2 + (voxA[2]-voxB[2])**2)
    

def watervoxel(vox, nearestwaters):
    watervox = (nearestwaters[0, vox[0], vox[1], vox[2]],
                 nearestwaters[1, vox[0], vox[1], vox[2]],
                 nearestwaters[2, vox[0], vox[1], vox[2]])
    return watervox

#distmatrix, voxelindex = graphshortestpaths(image, 0)
def graphshortestpaths(img, match):
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import shortest_path

    graph, voxelindex = constructgraph(img, match)
    graph = csr_matrix(graph)
    dist_matrix = shortest_path(csgraph=graph, directed=False, return_predecessors=False)
    return dist_matrix, voxelindex

def constructgraph(img, match):
    # get voxels in image that match a value
    nodes = [(i,j,k) for i in range(img.shape[0]) for j in range(img.shape[1]) for k in range(img.shape[2]) if img[(i,j,k)] <= match]
    
    # initialize adjacency matrix
    nnodes = len(nodes)
    adjmatrix = np.zeros((nnodes, nnodes))
    nodeindex = {}

    # get voxel numbering
    for i, vox in enumerate(nodes):
        nodeindex[vox] = i
	
    # fill in adjacency matrix, cubic grid rules (26 neighbors)
    for vox in nodeindex.keys():
        neighbors_all = [(vox[0] + offset[0], vox[1] + offset[1], vox[2] + offset[2]) for offset in neighbor_offsets]
        neighbors = [i for i in neighbors_all if i in nodeindex]
        for nb in neighbors:
            dist = euclideandistance(vox,nb)
            adjmatrix[nodeindex[vox], nodeindex[nb]] = dist
            adjmatrix[nodeindex[nb], nodeindex[vox]] = dist
            
    return adjmatrix, nodeindex
    



# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>

import argparse
import os.path
import os
import time
from xlms_tools.score import *
from xlms_tools.printing import *
from xlms_tools.visualization import *
from xlms_tools.simulate import outputSASDs

def main():
    parser = argparse.ArgumentParser(description='Score protein structures by their concordance with crosslinking mass spectrometry (XL-MS) data')
    parser.add_argument('pdb', metavar='PDB', type=str, nargs='+',
                    help='PDB file/s of protein: PDB or mmCIF format')
    parser.add_argument('-m', '--mode', type=str, choices=['score', 'depth', 'sim'], default='score',
                    help='score: compute model score with respect to XL-MS data; depth: compute residue depths; sim: simulate crosslinks and monolinks')
    parser.add_argument('-s','--score', type=str, choices=['old', 'new'], default='new',
                    help='old: use Total Residue Depth (RD) for monolinks; Matched and      Nonaccessible Crosslinks (MNXL) for crosslinks; \nnew: use Monolink Probability (MP) for monolinks, Crosslink Probability (XLP) for crosslinks')
    parser.add_argument('-l', '--list', type=str,
                    help='[score mode only] list of crosslinks and monolinks')
    parser.add_argument('-r', '--recov', type=str,
                    help='[sim mode only] recovery rate for crosslink/monolink simulation')
    parser.add_argument('--linker', type=str, choices=['BS3/DSS', 'DSBSO', 'PhoX'], 
                    default='BS3/DSS',
                    help='Crosslinking reagent used')
    parser.add_argument('--name', type=str, default='',
                    help='Run name. Output files will use this as a base filename')
    parser.add_argument('--color', type=str, default='lightgray',
                    help='color name or hexcode recognizable by ChimeraX, to color the protein in visualization')
    parser.add_argument('-q', '--quiet', action='store_true',
                    help='run with minimal stdout and output files (scores only, no visualization)')

    args = parser.parse_args()
    verbose = not args.quiet

    printsoftwareheader()

    if args.mode == 'score':
        print ('Scoring models...')
        start = time.time()
        
        # parse xl-ms file
        xllist = getxlmslist(args.list, args.linker)

        # set run name
        if len(args.name) > 0:
            visdir = args.name
        else:
            visdir = os.path.basename(args.list)[:-4]

        # if verbose, set up ChimeraX visulization
        if verbose:
            if not os.path.exists(visdir):
                os.mkdir(visdir)
            cxcptr = printcxc(args.pdb, visdir=visdir, color=args.color)

        # score each structure model
        bestmodel = ""
        bestscore = -1000
        for i,model in enumerate(args.pdb):
            struct = StructModel(model, verbose=verbose)
            struct.scoremodel(xllist, verbose=verbose)
            
            # open score output file
            if i == 0:
                f = openscorefile(struct, outname=visdir+"_scores")
                
            struct.printscores(f)
            if verbose:
                struct.printchimeraxfiles(visdir, cxcptr, xllist, i+1)

            # get best scoring model
            iscore = struct.avexlscore
            if iscore == 0:
                iscore = struct.avemlscore
            if iscore > bestscore:
                bestscore = iscore
                bestmodel = model
                
        f.close()   # close score output file
        
    
        print ("\n***\n\nBEST SCORING MODEL:", bestmodel)
        print (f"XLP/MP scores are in {visdir}_scores.tsv")
        if verbose:
            print (f"Open {visdir}.cxc with ChimeraX to visualize the models with crosslinks")
        print (f"Total time elapsed: {time.time()-start:.3f}s\n")    

    elif args.mode == 'depth':
        for i in args.pdb:
            struct = StructModel(i, verbose=verbose)
            struct.outputdepth()
        
    elif args.mode == 'sim':
        for i in args.pdb:
            outputSASDs(i, linker=args.linker)

    else:
        print("Mode '", args.mode, "' not yet supported")

    print("\nIf you use xlms-tools, please cite:")
    print("Manalastas-Cantos, K., Adoni, K. R., Pfeifer, M., Märtens, B., Grünewald, K., Thalassinos, K., & Topf, M. (2024). Modeling flexible protein structure with AlphaFold2 and cross-linking mass spectrometry. Molecular & Cellular Proteomics. https://doi.org/10.1016/j.mcpro.2024.100724\n")

if __name__ == '__main__':
    main()


# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>

from Bio.PDB import *
from xlms_tools.depth import *
from xlms_tools.parsers import *
from xlms_tools.visualization import *
import numpy as np
import time

class StructModel:
    def __init__(self, infile, depths=None, verbose=True):
        if verbose:
            print ("MODEL:", infile)

        # if no depths computed, we are starting from a pdb/cif file
        if depths == None:
            self.filename = infile
            self.biopdbstruct = parsebiopystructure(infile)
            self.residuedepths, self.residues, self.edt, self.nearestwaters, self.voxelized, self.atomlist, self.translatedatomcoords = computedepth(self.biopdbstruct, return_indices=True, verbose=verbose)

        # depth has been precomputed, probably api access
        else:
            self.filename = None
            self.biopdbstruct = infile
            self.residuedepths = depths
            self.residues, self.edt, self.nearestwaters = None, None, None
            self.voxelized, self.atomlist, self.translatedatomcoords = None, None, None
        
        self.maxCAdist = maxCAdistfromstructure(self.biopdbstruct)
        self.maxdepth = max(self.residuedepths.values())
        self.xlscores = {}
        self.avexlscore = 0
        self.avemlscore = 0
        self.depviols = 0
        self.distviols = 0
    
    def inmodel(self, link):
        found = False
        if link.__class__.__name__ == 'Monolink':
            if self.biopdbstruct[0][link.chainid][link.rnum].__contains__('CA'):
                return True
        else:   #if crosslink
            if self.biopdbstruct[0][link.siteA.chainid][link.siteA.rnum].__contains__('CA'):
                if self.biopdbstruct[0][link.siteB.chainid][link.siteB.rnum].__contains__('CA'):
                    return True
        return found
            
    def scoremodel(self, xlmslist, verbose=True):
        start = time.time()
        
        # accumulators for denominators
        totalmlweight = 0
        totalxlweight = 0
        xlweightbychain = {}
        
        # score each link
        for link in xlmslist:
            if self.inmodel(link):
                if link.__class__.__name__ == 'Monolink':
                    mlscore, depth, viol = link.score(self)
                    self.avemlscore += mlscore
                    totalmlweight += link.weight
                    if viol:
                        self.depviols += 1
                else:   # if crosslink
                    xlscore, dist, viol = link.score(self)
                    self.avexlscore += xlscore
                    totalxlweight += link.weight
                    chid = link.chains() 
                    if chid not in self.xlscores:
                        self.xlscores[chid] = xlscore
                        xlweightbychain[chid] = link.weight
                    else:
                        self.xlscores[chid] += xlscore
                        xlweightbychain[chid] += link.weight
                    if viol:
                        self.distviols += 1

        # get score averages
        if totalmlweight != 0:
            self.avemlscore /= totalmlweight
        if totalxlweight != 0:
            self.avexlscore /= totalxlweight
        for key in self.xlscores:
            self.xlscores[key] /= xlweightbychain[key]
        if verbose:
            print (f"-- Scoring model: {time.time()-start:.3f}s\n")    

    def printscores(self, ptr):
        
        # order output file columns
        intra = [i for i in self.xlscores if len(i) == 1]
        inter = [i for i in self.xlscores if len(i) == 2]
        intra.sort()
        inter.sort()
    
        
        ptr.write(f'{os.path.basename(self.filename)[:-4]}\t')
        if self.avemlscore != 0:
            ptr.write(f'{self.avemlscore:.3f}\t')
        else:
            ptr.write('--\t')
            
        
        for key in intra:
        	ptr.write(f'{self.xlscores[key]:.3f}\t')
        for key in inter:
        	ptr.write(f'{self.xlscores[key]:.3f}\t')
    
        ptr.write(f'{self.depviols}\t')
        ptr.write(f'{self.distviols}\t')
        ptr.write(f'{self.avexlscore:.3f}')            
        ptr.write('\n')
    
    def printchimeraxfiles(self, cxdir, cxcptr, xlmslist, modelid):
        # open pseudobond files
        g, m, b = createpb(self.filename, cxdir)
        
        # go through crosslinks and monolinks
        for link in xlmslist:
            if self.inmodel(link):
                if link.__class__.__name__ == 'Monolink':
                    # add visualization of monolink to ChimeraX command file
                    cxcptr.write(f'show #{modelid}/{link.chainid}:{link.rnum} atoms\n')
                    cxcptr.write(f'color #{modelid}/{link.chainid}:{link.rnum} {link.color(self)} atoms\n')
                    
                else:   # if crosslink
                    dist = link.eudist(self)
                    pos1 = f'{link.siteA.chainid}:{link.siteA.rnum}'
                    pos2 = f'{link.siteB.chainid}:{link.siteB.rnum}'
                    
                    if dist <= link.fiftypercent():
                        g.write(f'#{modelid}/{pos1}@ca #{modelid}/{pos2}@ca\n')
                    elif dist <= link.cutoff():
                        m.write(f'#{modelid}/{pos1}@ca #{modelid}/{pos2}@ca\n')
                    else:
                        b.write(f'#{modelid}/{pos1}@ca #{modelid}/{pos2}@ca\n')

        # close pseudobond files
        g.close()
        m.close()
        b.close()
                    
    def outputdepth(self):
        with open(self.biopdbstruct.id+'-residue.depth', 'w') as f:
            for key in self.residuedepths.keys():
                f.write(f'{key}\t{self.residues[key]}\t{self.residuedepths[key]}\n')    
    
class Crosslink:
    def __init__(self, monolinkA, monolinkB, weight=1.0):
        self.siteA = monolinkA
        self.siteB = monolinkB
        self.weight = float(weight)

    def linkdetails(self):
        self.siteA.linkdetails()
        self.siteB.linkdetails()
        print (self.weight)
    
    def chains(self):
        if self.siteA.chainid == self.siteB.chainid:
            return self.siteA.chainid
        else:
            return self.siteA.chainid+self.siteB.chainid

    def probability(self, dist):
        if self.siteA.linker == 'BS3/DSS':    # default situation, BS3 or DSS
            return 1/(1+np.exp(0.33*dist-7))    # this is the best from the benchmark
        elif self.siteA.linker == 'DSBSO':
            return logistic_function(dist, 1.07, 0.18, 20.25)
        elif self.siteA.linker == 'PhoX':
            return logistic_function(dist, 1.05, 0.34, 15.50)
        else:
            print ('No information about linker. Using distance distribution for BS3/DSS')
            return 1/(1+np.exp(0.33*dist-7))

    def cutoff(self):
        cutoffs = {'BS3/DSS':33, 'DSBSO':40, 'PhoX':30}
        if self.siteA.linker in cutoffs:
            return cutoffs[self.siteA.linker]
        else:
            print ('No information about linker. Using distance cutoff for BS3/DSS')
            return cutoffs['BS3/DSS']
                
    def eudist(self, model):
        ca1 = model.biopdbstruct[0][self.siteA.chainid][self.siteA.rnum]['CA']
        ca2 = model.biopdbstruct[0][self.siteB.chainid][self.siteB.rnum]['CA']
        return ca1-ca2
    
    def fiftypercent(self):
        cutoffs = {'BS3/DSS':21, 'DSBSO':21, 'PhoX':16}
        if self.siteA.linker in cutoffs:        
            return cutoffs[self.siteA.linker]  
        else:
            print ('No information about linker. Using distance cutoff for BS3/DSS')
            return cutoffs['BS3/DSS']
         
    def score(self, model):
        viol = False
        dep1 = self.siteA.depth(model)
        dep2 = self.siteB.depth(model)
        mp1 = self.siteA.probability(dep1)
        mp2 = self.siteB.probability(dep2)
        dist = self.eudist(model)
        distp = self.probability(dist)
        score = mp1*mp2*distp
        
        if dep1 >= self.siteA.cutoff():
            score = -dep1/(4*model.maxdepth)
        if dep2 >= self.siteB.cutoff():
            score = -dep2/(4*model.maxdepth)
        if dist > self.cutoff():
            score = -dist/(2*model.maxCAdist)
            viol = True
        score *= self.weight
        return score, dist, viol
        
class Monolink:
    def __init__(self, chainid, rnum, linker='BS3/DSS', weight=1.0):
        self.chainid = chainid
        self.rnum = int(rnum)
        self.linker = linker
        self.weight = float(weight)
    
    def linkdetails(self):
        print (self.chainid, self.rnum, self.linker, self.weight)
            
    def probability(self, depth):
        if self.linker == 'BS3/DSS':
            tagcoefs = np.array([1.20875756e+03, 5.31961163e+00, 6.43465155e-04])
            lyscoefs = np.array([4.44470936e+02, 4.65135510e+00, 1.69863402e-03])
            if func(depth, *lyscoefs) > 0:
                return func(depth, *tagcoefs)/func(depth, *lyscoefs)
            else:
                return 0
        elif self.linker == 'DSBSO':
            tagcoefs = np.array([2583, 5.38])
            lyscoefs = np.array([318, 3.94])
            if func2(depth, *lyscoefs) > 0:
                return func2(depth, *tagcoefs)/func2(depth, *lyscoefs)
            else:
                return 0
        elif self.linker == 'PhoX':
            tagcoefs = np.array([2896, 5.47])
            lyscoefs = np.array([288, 3.88])
            if func2(depth, *lyscoefs) > 0:
                return func2(depth, *tagcoefs)/func2(depth, *lyscoefs)
            else:
                return 0

    def cutoff(self):
        cutoffs = {'BS3/DSS':15, 'DSBSO':15, 'PhoX':17}
        if self.linker in cutoffs:        
            return cutoffs[self.linker]  
        else:
            print ('No information about linker. Using depth cutoff for BS3/DSS')
            return cutoffs['BS3/DSS']
        
    def depth(self, model):
        return model.residuedepths[f'{self.chainid}:{self.rnum}']
    
    def color(self, model):
        depth = self.depth(model)
        color = 'blue'
        
        if self.linker == 'BS3/DSS':
            if depth > 6.25:
                color = 'red'
        elif self.linker == 'DSBSO':
            if depth > 6.25:
                color = 'red'
        elif self.linker == 'PhoX':
            if depth > 6.25:
                color = 'red'
        else:
            if depth > 6.25:
                color = 'red'            
        return color
        
    def score(self, model):
        viol = False
        depth = self.depth(model) 
        if depth >= self.cutoff():
            score = -(depth/model.maxdepth)
            viol = True
        else:
            score = self.probability(depth)
        score *= self.weight
        return score, depth, viol
        
def getxlmslist(xlmslist, linker='BS3/DSS'):
    xllist = []
    unique = []
    with open(xlmslist, 'r') as f:
        for ln in f:
            link, linktuple = parsexlmsline(ln.strip(), linker=linker)
            if linktuple not in unique:
                unique.append(linktuple)
                xllist.append(link)
    return xllist

def parsexlmsline(line, linker='BS3/DSS', weight=1.0):
    tmp = line.split()
    if len(tmp) > 1:    #if there's a score
        weight = float(tmp[1])
        buf = tmp[0].split('|')
    else:
        buf = line.split('|')
        
    if len(buf) == 2:   # if monolink
        res1 = int(buf[0])
        ch1 = buf[1]
        site1 = Monolink(ch1, res1, linker=linker, weight=weight)
        return site1, (ch1, res1)
        
    elif len(buf) == 4:   # if crosslink
        res1 = int(buf[0])
        ch1 = buf[1]
        site1 = Monolink(ch1, res1, linker=linker)
        res2 = int(buf[2])
        ch2 = buf[3]
        site2 = Monolink(ch2, res2, linker=linker)
        if (ch1, res1) <= (ch2, res2):
            return Crosslink(site1, site2, weight=weight), (ch1, res1, ch2, res2)
        else:
            return Crosslink(site2, site1, weight=weight), (ch2, res2, ch1, res1)

def linkmetrics(structure, linktuple, depths, linkweight=1.0, linker='BS3/DSS'):
    # initialize structure model
    model = StructModel(structure, depths=depths)
    
    # initialize crosslink or monolink
    if len(linktuple) == 2:  #if a monolink
        link = Monolink(linktuple[1], linktuple[0], linker=linker, weight=linkweight)
    else:   #if a crosslink
        ml1 = Monolink(linktuple[1], linktuple[0], linker=linker)
        ml2 = Monolink(linktuple[3], linktuple[2], linker=linker)
        link = Crosslink(ml1, ml2, weight=linkweight)
    
    # compute score and depth/distance
    if model.inmodel(link): # link found in model
        score, d, viol = link.score(model)
        return score, d
    else:   # link not found
        print ("Link coordinates not found in model")
        return None, None
            
def func(x, a, b, c):
    return (a * (x**-b)) + c

def func2(x, a, b):
    return (a * (x**-b))

def logistic_function(x, L, k, x0):
    return L / (1 + np.exp(k * (x - x0)))




# This file is part of the xlms-tools package
#
# Copyright (c) 2023 - Topf Lab, Leibniz-Institut für Virologie
# and Center for Data and Computing in Natural Sciences, Universität Hamburg
# Hamburg, Germany.
#
# This module was developed by:
#   Karen Manalastas-Cantos    <karen.manalastas-cantos AT cssb-hamburg.de>
